#include <math.h>
#include "amx.h"

#ifndef PI
  #define PI  3.1415926535897932384626433832795
#endif

// CORDIC֐JԂ
#define COR_ITE_NUM 29

// CORDIC֐pW
int g_iCordicTable[COR_ITE_NUM];
float g_fPos2BinScale;
float g_fBin2PosScale;
float g_fAng2BinScale;
float g_fBin2AngScale;
float g_fPosLimit;
int   g_fKFactor;

// CORDIC֐vg^Cv
void  cordic_init();
float cordic_atan(float Y, float X);
void  cordic_cos_sin(float angle, float *cor_cos, float *cor_sin);

// {bgA[
float g_fLength1 = 1.0;
float g_fLength2 = 1.0;

static cell AMX_NATIVE_CALL n_SetArmLength(AMX *amx,const cell *params)
{
	g_fLength1 = amx_ctof(params[1]);
	g_fLength2 = amx_ctof(params[2]);
	return 0;
}

static cell AMX_NATIVE_CALL n_Kinematics(AMX *amx,const cell *params)
{
	cell *cPtr1 , *cPtr2 , ret = (-1);
	int  res1 , res2;

#if defined PAWN4
	// Pawno[W4n̏ꍇ
	cPtr1 = amx_Address(amx , params[1] );		// PawnVM̃AhXC̃AhX֕ϊ
	cPtr2 = amx_Address(amx , params[2] );
	if( (cPtr1 == NULL) || (cPtr2 == NULL) ) {
		// error
		return ret;
	}
#else
	// o[W4ȑȌꍇ
	res1 = amx_GetAddr(amx, params[1], &cPtr1);	// PawnVM̃AhXC̃AhX֕ϊ
	res2 = amx_GetAddr(amx, params[2], &cPtr2);
	if( (res1 != AMX_ERR_NONE) || (res2 != AMX_ERR_NONE) ) {
		// error
		return ret;
	}
#endif
	float angle1 = amx_ctof(cPtr1[0]);
	float angle2 = amx_ctof(cPtr1[1]);
	float pos[2];

	float C1 , S1 , C12 , S12;
	float Tc;

	// pxPʂ̕ϊ@i3Ŏwj
	switch( params[3] ){
		case 0:		// radianP
			Tc = 180.0 / PI;
			angle1 = angle1 * Tc;
			angle2 = angle2 * Tc;
			break;
		case 2:		// gradP
			Tc = 9.0 / 10.0;
			angle1 = angle1 * Tc;
			angle2 = angle2 * Tc;
			break;
		default:	// degreeP
			break;
	}

	// ʒǔvZ
	cordic_cos_sin( angle1 , &C1 , &S1 );
	cordic_cos_sin( (angle1+angle2) , &C12 , &S12 );
	pos[0] = g_fLength1 * C1 + g_fLength2 * C12 ;
	pos[1] = g_fLength1 * S1 + g_fLength2 * S12 ;

	cPtr2[0] = amx_ftoc(pos[0]);
	cPtr2[1] = amx_ftoc(pos[1]);

	ret = 1;
	return ret;
}

static cell AMX_NATIVE_CALL n_inverseKinematics(AMX *amx,const cell *params)
{
	cell *cPtr1 , *cPtr2 , ret = (-1);
	int  res1 , res2;

#if defined PAWN4
	// Pawno[W4n̏ꍇ
	cPtr1 = amx_Address(amx , params[1] );		// PawnVM̃AhXC̃AhX֕ϊ
	cPtr2 = amx_Address(amx , params[2] );
	if( (cPtr1 == NULL) || (cPtr2 == NULL) ) {
		// error
		return ret;
	}
#else
	// o[W4ȑȌꍇ
	res1 = amx_GetAddr(amx, params[1], &cPtr1);	// PawnVM̃AhXC̃AhX֕ϊ
	res2 = amx_GetAddr(amx, params[2], &cPtr2);
	if( (res1 != AMX_ERR_NONE) || (res2 != AMX_ERR_NONE) ) {
		// error
		return ret;
	}
#endif
	float c2 , s2;
	float angle1=0.0 , angle2=0.0;
	float px = amx_ctof(cPtr1[0]);	// xW
	float py = amx_ctof(cPtr1[1]);	// yW
	int   select = (int)params[3];	// ̑I
	float Tc;

	c2 = ( px*px + py*py - g_fLength1*g_fLength1 - g_fLength2*g_fLength2)/(2.0*g_fLength1*g_fLength2);
	// Bsȏꍇ̓G[
	if(c2 > 1.0)    return ret;
	else if(c2 < -1.0) return ret;

	s2 = sqrtf(1.0 - c2*c2);

	// ֐ߊpx̌vZ
	switch(select){
		case 1:
			angle1 = cordic_atan( py, px ) - cordic_atan( g_fLength2*s2 , g_fLength1 + g_fLength2*c2 );
			angle2 = cordic_atan( s2 , c2);
			break;
		case 2:
			angle1 = cordic_atan( py , px) + cordic_atan( g_fLength2*s2 , g_fLength1 + g_fLength2*c2 );
			angle2 = cordic_atan(-s2 , c2 );
			break;
	}

	// pxPʂ̕ϊ@i4Ŏwj
	switch( params[4] ){
		case 0:		// radianP
			Tc = PI / 180.0;
			angle1 = angle1 * Tc;
			angle2 = angle2 * Tc;
			break;
		case 2:		// gradP
			Tc = 10.0 / 9.0;
			angle1 = angle1 * Tc;
			angle2 = angle2 * Tc;
			break;
		default:	// degreeP
			break;
	}

	cPtr2[0] = amx_ftoc( angle1 );
	cPtr2[1] = amx_ftoc( angle2 );

	ret = 1;
	return ret;
}

static cell AMX_NATIVE_CALL n_CordicCosSin(AMX *amx,const cell *params)
{
	cell *cPtr , *sPtr , ret = (-1);
	int  res1 , res2;

#if defined PAWN4
	// Pawno[W4n̏ꍇ
	cPtr = amx_Address(amx , params[2] );		// PawnVM̃AhXC̃AhX֕ϊ
	sPtr = amx_Address(amx , params[3] );
	if( (cPtr1 == NULL) || (cPtr2 == NULL) ) {
		// error
		return ret;
	}
#else
	// o[W4ȑȌꍇ
	res1 = amx_GetAddr(amx, params[2], &cPtr);	// PawnVM̃AhXC̃AhX֕ϊ
	res2 = amx_GetAddr(amx, params[3], &sPtr);
	if( (res1 != AMX_ERR_NONE) || (res2 != AMX_ERR_NONE) ) {
		// error
		return ret;
	}
#endif

	float angle = amx_ctof( params[1] );
	switch( params[4] ){
		case 0:		// radianP
			angle = angle * 180.0 / PI;
			break;
		case 2:		// gradP
			angle = angle * 9.0 / 10.0;
			break;
		default:	// degreeP
			break;
	}
	cordic_cos_sin( angle, (float *)cPtr , (float *)sPtr );
	return 1;
}

static cell AMX_NATIVE_CALL n_CordicAtan2(AMX *amx,const cell *params)
{
	float y = amx_ctof( params[1] );
	float x = amx_ctof( params[2] );
	float angle;
	
	angle = cordic_atan( y, x );

	switch( params[3] ){
		case 0:		// radianP
			angle = angle * PI / 180.0;
			break;
		case 2:		// gradP
			angle = angle * 10.0 / 9.0;
			break;
		default:	// degreeP
			break;
	}

	return amx_ftoc(angle);
}

static cell AMX_NATIVE_CALL n_floatatan2(AMX *amx,const cell *params)
{
	float y = amx_ctof( params[1] );
	float x = amx_ctof( params[2] );
	float angle;
	
	angle = atan2( y, x );
	
	switch( params[3] ){
		case 1:		// degreeP
			angle = angle * 180.0 / PI;
			break;
		case 2:		// gradP
			angle = angle * 200.0 / PI;
			break;
		default:	// radianP
			break;
	}
	
	return amx_ftoc(angle);
}

static cell AMX_NATIVE_CALL n_floatcosf(AMX *amx,const cell *params)
{
	float angle = amx_ctof( params[1] );
	float Tc;
	float f;

	// pxPʂ̕ϊ@i2Ŏwj
	switch( params[2] ){
		case 1:		// degreeP
			Tc = PI / 180.0;
			angle = angle * Tc;
			break;
		case 2:		// gradP
			Tc = PI / 200.0;
			angle = angle * Tc;
			break;
		default:	// radianP
			break;
	}

	f= cosf(angle);
	return amx_ftoc(f);
}

static cell AMX_NATIVE_CALL n_floatsinf(AMX *amx,const cell *params)
{
	float angle = amx_ctof( params[1] );
	float Tc;
	float f;

	// pxPʂ̕ϊ@i2Ŏwj
	switch( params[2] ){
		case 1:		// degreeP
			Tc = PI / 180.0;
			angle = angle * Tc;
			break;
		case 2:		// gradP
			Tc = PI / 200.0;
			angle = angle * Tc;
			break;
		default:	// radianP
			break;
	}

	f= sinf(angle);
	return amx_ftoc(f);
}

static cell AMX_NATIVE_CALL n_floatatan2f(AMX *amx,const cell *params)
{
	float y = amx_ctof( params[1] );
	float x = amx_ctof( params[2] );
	float angle;
	
	angle = atan2f( y, x );
	
	switch( params[3] ){
		case 1:		// degreeP
			angle = angle * 180.0 / PI;
			break;
		case 2:		// gradP
			angle = angle * 200.0 / PI;
			break;
		default:	// radianP
			break;
	}
	
	return amx_ftoc(angle);
}

static cell AMX_NATIVE_CALL n_floatsqrootf(AMX *amx,const cell *params)
{
	float val = amx_ctof(params[1]);
	float f;

	if (val < 0) return amx_RaiseError(amx, AMX_ERR_DOMAIN);
	f = sqrtf(val);
	return amx_ftoc(f);
}


int amx_ArmExtensionInit(AMX *amx)
{
	static AMX_NATIVE_INFO Arm_Extension[] = 
	{
		{ "SetArmLength"      , n_SetArmLength },
		{ "Kinematics"        , n_Kinematics },
		{ "inverseKinematics" , n_inverseKinematics },
		{ "floatatan2"        , n_floatatan2 },
		{ "floatcosf"         , n_floatcosf },
		{ "floatsinf"         , n_floatsinf },
		{ "floatatan2f"       , n_floatatan2f },
		{ "floatsqrootf"      , n_floatsqrootf },
		{ "CordicCosSin"      , n_CordicCosSin },
		{ "CordicAtan2"       , n_CordicAtan2 },
		{ NULL    , NULL }        // terminator
	};
	cordic_init();

	return amx_Register(amx , Arm_Extension , -1);
}

int amx_ArmExtensionCleenup(AMX *amx){
	return AMX_ERR_NONE;
}

void cordic_init()
{
	g_fBin2PosScale = 1.0/(float)(1 << (COR_ITE_NUM - 3) );
	g_fPos2BinScale = (float)(1 << (COR_ITE_NUM - 3) );
	g_fBin2AngScale = 180.0/(float)(1 << (COR_ITE_NUM + 1) );
	g_fAng2BinScale = (float)(1 << (COR_ITE_NUM + 1) )/180.0;
	g_fPosLimit     = (float)( 0x20000000 )/g_fPos2BinScale;
/*
	// JԂ񐔁iCOR_ITE_NUMjύXꍇarctañe[uf[^Ɖ][h̏lvZ
	int i;
	double angle;
	double KScale;
	
	KScale = 1.0;
	for( i=0 ; i<COR_ITE_NUM ; i++) {
		angle = atan(1.0/(double)(1<<i));
		g_iCordicTable[i] = (int)round((180.0*angle/PI)*((float)(1 << (COR_ITE_NUM+1) )/180.0)) ;
		KScale *= cos(angle);
	}
	g_fKFactor = (int)round( KScale*g_fPos2BinScale );
*/
	// arctan( 1/(2^i) )̃e[u
	g_iCordicTable[ 0]  = 0x10000000;
	g_iCordicTable[ 1]  = 0x0972028F;
	g_iCordicTable[ 2]  = 0x04FD9C2E;
	g_iCordicTable[ 3]  = 0x028888EA;
	g_iCordicTable[ 4]  = 0x014586A2;
	g_iCordicTable[ 5]  = 0x00A2EBF1;
	g_iCordicTable[ 6]  = 0x00517B0F;
	g_iCordicTable[ 7]  = 0x0028BE2B;
	g_iCordicTable[ 8]  = 0x00145F2A;
	g_iCordicTable[ 9]  = 0x000A2F97;
	g_iCordicTable[10]  = 0x000517CC;
	g_iCordicTable[11]  = 0x00028BE6;
	g_iCordicTable[12]  = 0x000145F3;
	g_iCordicTable[13]  = 0x0000A2FA;
	g_iCordicTable[14]  = 0x0000517D;
	g_iCordicTable[15]  = 0x000028BE;
	g_iCordicTable[16]  = 0x0000145F;
	g_iCordicTable[17]  = 0x00000A30;
	g_iCordicTable[18]  = 0x00000518;
	g_iCordicTable[19]  = 0x0000028C;
	g_iCordicTable[20]  = 0x00000146;
	g_iCordicTable[21]  = 0x000000A3;
	g_iCordicTable[22]  = 0x00000051;
	g_iCordicTable[23]  = 0x00000029;
	g_iCordicTable[24]  = 0x00000014;
	g_iCordicTable[25]  = 0x0000000A;
	g_iCordicTable[26]  = 0x00000005;
	g_iCordicTable[27]  = 0x00000003;
	g_iCordicTable[28]  = 0x00000001;
	// ][h(cos,siňvZ)̏l
	g_fKFactor = 0x026DD3B7;

}

float cordic_atan(float Y, float X)
{
	int i;
	int int_X , int_Y ,tmp;
	int x0 , y0 , z0, x1 , y1 , z1;
	float cor_atan;

	float xt , yt;
	
	xt = X;
	yt = Y;
	
	// Zsۂ͈͓̔Ɏ܂悤ɃXP[O
	while( (fabs(xt) > g_fPosLimit ) || (fabs(yt) > g_fPosLimit ) ) {
		xt = xt / 2.0;
		yt = yt / 2.0;
	}

	int_X = (int)round(xt*g_fPos2BinScale);
	int_Y = (int)round(yt*g_fPos2BinScale);

	// 1یȊǑvZs
	if (int_X < 0) {
		tmp = int_X;
		if (int_Y > 0) {
			x0 = int_Y;
			y0 = -tmp;
			z0 = (int)round(90.0*g_fAng2BinScale);
		} else {
			x0 = -int_Y;
			y0 = tmp;
			z0 = 0-(int)round(90.0*g_fAng2BinScale);
		}
	} else {
  	  	x0 = int_X;
  	  	y0 = int_Y;
		z0 = 0;
	}

	// CORDIC Vector[hZ
	for (i = 0; i < COR_ITE_NUM ; i++) {
		if (y0 > 0) {
			x1 = x0 + (y0 >> i);
			y1 = y0 - (x0 >> i);
			z1 = z0 + g_iCordicTable[i];
		} else {
			x1 = x0 - (y0 >> i);
			y1 = y0 + (x0 >> i);
			z1 = z0 - g_iCordicTable[i];
		}
		x0 = x1;
		y0 = y1;
		z0 = z1;
	}

	// Zʂfloat^(Pʂdegree)ɕϊ
	cor_atan = (float)(z1)*g_fBin2AngScale;
	if(cor_atan > 180.0) cor_atan = 180.0;
	else if(cor_atan < -180.0) cor_atan = -180.0;
	
	return cor_atan;
}

void cordic_cos_sin(float angle, float *cor_cos, float *cor_sin)
{
	int i , flag;
	long x0 , y0 , z0, x1 , y1 , z1 , int_angle;
	float tmp_angle;

	flag = 0;

	tmp_angle = angle;

	// 1یȊǑvZs
	if(tmp_angle < (-90.0)){
		flag =1;
		tmp_angle += 180.0;
	}
	else if(tmp_angle > 90.0){
		flag =1;
		tmp_angle -= 180.0;
	}else{
		flag = 0;
	}

	int_angle = (long)round(tmp_angle*g_fAng2BinScale);
	if(int_angle<0){
		int_angle = 0 - int_angle;
		flag |= 0x2;
	}

	// CORDIC Rotation[hZ
	x0 = g_fKFactor;
	y0 = 0;
	z0 = int_angle;

	for(i=0 ; i<COR_ITE_NUM ; i++){
		if(z0 >= 0){
			x1 = x0 - (y0>>i);
			y1 = y0 + (x0>>i);
			z1 = z0 - g_iCordicTable[i];
		}
		else{
			x1 = x0 + (y0>>i);
			y1 = y0 - (x0>>i);
			z1 = z0 + g_iCordicTable[i];
		}
		x0 = x1;
		y0 = y1;
		z0 = z1;
	}

	// Zʂfloat^ɕϊ
	*cor_cos = (float)(x1) * g_fBin2PosScale;
	*cor_sin = (float)(y1) * g_fBin2PosScale;

	// 1یȊǑvZsꍇ̌㏈
	if((flag & 0x2)==2){
		*cor_sin = 0-(*cor_sin);
	}
	if((flag & 0x1)==1){
		*cor_cos = 0-(*cor_cos);
		*cor_sin = 0-(*cor_sin);
	}
}
