#include "pawn_lm3s3748_IO.h"

#define DIGITAL_IO_MAX 16

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_ints.h"
#include "driverlib/sysctl.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"

static const unsigned long g_ulSYSCTL_PERIPH_GPIO[8] =
{
	SYSCTL_PERIPH_GPIOA, SYSCTL_PERIPH_GPIOB, SYSCTL_PERIPH_GPIOC, SYSCTL_PERIPH_GPIOD, 
	SYSCTL_PERIPH_GPIOE, SYSCTL_PERIPH_GPIOF, SYSCTL_PERIPH_GPIOG, SYSCTL_PERIPH_GPIOH
};

static const unsigned long g_ulGPIOBase[8] =
{
	GPIO_PORTA_BASE, GPIO_PORTB_BASE, GPIO_PORTC_BASE, GPIO_PORTD_BASE,
	GPIO_PORTE_BASE, GPIO_PORTF_BASE, GPIO_PORTG_BASE, GPIO_PORTH_BASE
};

static const unsigned long g_ulGPIOInt[8] =
{ 
	INT_GPIOA, INT_GPIOB, INT_GPIOC, INT_GPIOD, INT_GPIOE, INT_GPIOF, INT_GPIOG, INT_GPIOH
};

unsigned char g_NumDigitalOut = 0;
unsigned char g_NumDigitalIn  = 0;

struct GPIO_PIN_INFO {
	unsigned long port;		/* |[g */
	unsigned char pin;		/* sԍ */
};

struct PIO_INFO {
	struct GPIO_PIN_INFO Pins[DIGITAL_IO_MAX];
};

struct PIO_INFO g_DO;
struct PIO_INFO g_DI;

int g_GPIOIntHandlerId[8];	// Pawn̊荞݊֐ID
tBoolean g_GPIOPortIntRegistered[8] = {false , false , false , false , false , false , false , false}; 	// GPIOPort荞ݓo^tO
tBoolean g_GPIOPinIntRegistered[DIGITAL_IO_MAX]; // GPIOPin荞ݓo^tO

// GPIO荞݊֐vg^Cv
void PORTAIntHandler(void);
void PORTBIntHandler(void);
void PORTCIntHandler(void);
void PORTDIntHandler(void);
void PORTEIntHandler(void);
void PORTFIntHandler(void);
void PORTGIntHandler(void);
void PORTHIntHandler(void);
// 荞݊֐z
void (*GPIOIntHandler[])(void) = { 	PORTAIntHandler , PORTBIntHandler , PORTCIntHandler , PORTDIntHandler ,
									PORTEIntHandler , PORTFIntHandler , PORTGIntHandler , PORTHIntHandler};

int amx_putstr(const char* s);

static cell AMX_NATIVE_CALL n_PinTypeGPIOOutput(AMX *amx, const cell *params)
{
	unsigned long pf_code = ( unsigned long )params[1];				// IOR[h(1)̎擾
	unsigned char pin_code = ( unsigned char )( pf_code >> 24 );	// IOR[h̏8rbg擾
	unsigned long ulPort;											// GPIO|[g̃x[XEAhX
	unsigned char ucPins;											// sE}XN
	unsigned long ulStrength = (unsigned long)params[2];			// o͓d(2j̎擾
	unsigned char port_no;
	unsigned long func_code;
	cell ret = -1;

	if( ( ( pf_code >> 8 ) & 0xFFFF )== 0 ) {		// w肳ꂽsɋ@\蓖ĂĂȂꍇ
		if( g_NumDigitalOut < DIGITAL_IO_MAX ) {
			port_no = (pin_code >> 4) & 0x7;					// |[gԍ̎擾
			ulPort = g_ulGPIOBase[ port_no ];					// sԍ̎擾
			ucPins = ( 1 << (pin_code & 0x7) );					// s}XN쐬
			func_code = PinFuncDigitalOut | g_NumDigitalOut;	// @\R[h̐ݒ

			if( SetPinFuncCode( amx, pin_code , func_code ) == AMX_ERR_NONE ) {		// IOR[hPawnVM֏߂
				MAP_SysCtlPeripheralEnable( g_ulSYSCTL_PERIPH_GPIO[ port_no ] );	// w肵GPIOytFLɂ
				MAP_GPIODirModeSet( ulPort , ucPins , GPIO_DIR_MODE_OUT);			// w肵sGPIOo͂ƂĐݒ
				MAP_GPIOPadConfigSet( ulPort , ucPins , ulStrength , GPIO_PIN_TYPE_STD );

				g_DO.Pins[ g_NumDigitalOut ].port = ulPort;		// |[g̃AhXۑ
				g_DO.Pins[ g_NumDigitalOut ].pin  = ucPins;		// sE}XNۑ
				ret = (pin_code << 24) | func_code ;			// IOR[h߂l֑
				g_NumDigitalOut++;								// ݒςGPIOo͐CNg
			}
		}
	}

	return ret;
}

static cell AMX_NATIVE_CALL n_PinTypeGPIOInput(AMX *amx, const cell *params)
{
	unsigned long pf_code = ( unsigned long )params[1];				// IOR[h(1)̎擾
	unsigned char pin_code = ( unsigned char )( pf_code >> 24 );	// IOR[h̏8rbg擾
	unsigned long ulPort;											// GPIO|[g̃x[XEAhX
	unsigned char ucPins;											// sE}XN
	unsigned long ulPinType = (unsigned long)params[2];				// sE^Cv(2j̎擾
	unsigned char port_no;
	unsigned long func_code;
	cell ret = -1;

	if( ( ( pf_code >> 8 ) & 0xFFFF )== 0 ) {	// w肳ꂽsɋ@\蓖ĂĂȂꍇ
		if( g_NumDigitalIn < DIGITAL_IO_MAX ) {
			port_no = (pin_code >> 4) & 0x7;					// |[gԍ̎擾
			ulPort = g_ulGPIOBase[ port_no ];					// sԍ̎擾
			ucPins = ( 1 << (pin_code & 0x7) );					// s}XN쐬
			func_code = PinFuncDigitalIn | g_NumDigitalIn;		// @\R[h̐ݒ

			if( SetPinFuncCode( amx, pin_code , func_code ) == AMX_ERR_NONE ) {		// IOR[hPawnVM֏߂
				MAP_SysCtlPeripheralEnable( g_ulSYSCTL_PERIPH_GPIO[ port_no ] );	// w肵GPIOytFLɂ
				MAP_GPIODirModeSet( ulPort , ucPins , GPIO_DIR_MODE_IN);			// w肵sGPIO͂ƂĐݒ
				MAP_GPIOPadConfigSet( ulPort , ucPins , GPIO_STRENGTH_2MA, ulPinType );
			
				g_DI.Pins[ g_NumDigitalIn ].port = ulPort;		// |[g̃AhXۑ
				g_DI.Pins[ g_NumDigitalIn ].pin  = ucPins;		// sE}XNۑ
				ret = (pin_code << 24) | func_code ;			// IOR[h߂l֑
				g_NumDigitalIn++;								// ݒςGPIO͐CNg
			}
		}
	}

	return ret;
}

static cell AMX_NATIVE_CALL n_GPIOWrite(AMX *amx,const cell *params)
{
	unsigned long pf_code = (unsigned long )params[1];		// IOR[h(1)̎擾
	unsigned char flag  = ( unsigned char )params[2];		// o͒l(2j̎擾
	unsigned char dev_no;
	int done = -1;

	// w肳ꂽsGPIOo͂̏ꍇ́Cs̏Ԃݒ
	if( ( pf_code & PinFuncDigitalOut ) == PinFuncDigitalOut ) {
		dev_no = pf_code & 0xF;		// GPIOo͔ԍ擾
		if( dev_no < g_NumDigitalOut ) {
			if (flag) {
				MAP_GPIOPinWrite(g_DO.Pins[ dev_no ].port , g_DO.Pins[ dev_no ].pin , g_DO.Pins[dev_no].pin);
			}
			else {
				MAP_GPIOPinWrite(g_DO.Pins[ dev_no ].port , g_DO.Pins[ dev_no ].pin , 0);
			}
			done = 1;
		}
	}
	if( done == (-1) ) amx_putstr("illegal I/O\n");

	return 0;
}

static cell AMX_NATIVE_CALL n_GPIORead(AMX *amx,const cell *params)
{
	unsigned long pf_code = (unsigned long )params[1];			// IOR[h(1)̎擾
	unsigned char dev_no;
	cell ret = -1;

	// w肳ꂽsGPIOo()̏ꍇ́Cs̏Ԃǂݏo
	dev_no = pf_code & 0xF;
	if( ( pf_code & PinFuncDigitalIn ) == PinFuncDigitalIn ) {	//* GPIO͂̏ꍇ
		if( dev_no < g_NumDigitalIn ) {
			ret = MAP_GPIOPinRead(g_DI.Pins[ dev_no ].port , g_DI.Pins[ dev_no ].pin) == g_DI.Pins[ dev_no ].pin;
		}
	}
	else if( ( pf_code & PinFuncDigitalOut ) == PinFuncDigitalOut ) {	// GPIOo͂̏ꍇ
		if( dev_no < g_NumDigitalOut ) {
			ret = MAP_GPIOPinRead(g_DO.Pins[ dev_no ].port , g_DO.Pins[ dev_no ].pin) == g_DO.Pins[ dev_no ].pin;
		}
	}
	if( ret == (-1) ) amx_putstr("illegal I/O\n");

	return ret;
}

static cell AMX_NATIVE_CALL n_GPIOPinIntEnable(AMX *amx,const cell *params)
{
	unsigned long pf_code = ( unsigned long )params[1];				// IOR[h(1)̎擾
	unsigned char pin_code = ( unsigned char )( pf_code >> 24 );	// IOR[h̏8rbg擾
	unsigned long ulIRQType = (unsigned long)params[2];				// 荞݃^Cv(2)̎擾
	unsigned long ulPort;
	unsigned char ucPins;
	unsigned long func_code = pf_code & 0xFFFFFF;
	unsigned char dev_no;
	cell ret = -1;
	
	if( ( pf_code & PinFuncDigitalIn ) == PinFuncDigitalIn ) {	// GPIO͂̏ꍇ
		dev_no = pf_code & 0xF;
		if( dev_no < g_NumDigitalIn ) {
			func_code |= PinFuncDigitalIrq;		// @\R[h̐ݒ(荞ݗL)

			if( SetPinFuncCode( amx , pin_code , func_code ) == AMX_ERR_NONE ) {		// IOR[hPawnVM֏߂
				ulPort = g_DI.Pins[ dev_no ].port;	// |[g̃x[XAhX擾
				ucPins = g_DI.Pins[ dev_no ].pin;	// sE}XN擾
	
				MAP_GPIOPinIntDisable( ulPort , ucPins );			// GPIOPin荞݂𖳌ɂ
				MAP_GPIOIntTypeSet( ulPort , ucPins , ulIRQType );	// 荞݃^Cv̐ݒ
				MAP_GPIOPinIntEnable( ulPort , ucPins );			// GPIOPin荞݂Lɂ
				ret = (pin_code << 24) | func_code ;				// IOR[h߂l֑
				g_GPIOPinIntRegistered[ dev_no ] = true;			// GPIOPin荞ݓo^tO𗧂Ă
			}
		}
	}
	if( ret == (-1) ) amx_putstr("illegal I/O\n");

	return ret;
}

static cell AMX_NATIVE_CALL n_GPIOPinIntDisable(AMX *amx,const cell *params)
{
	unsigned long pf_code = ( unsigned long )params[1];				// IOR[h(1)̎擾
	unsigned char pin_code = ( unsigned char )( pf_code >> 24 );	// IOR[h̏8rbg擾
	unsigned long pf_mask = ( PinFuncDigitalIn | PinFuncDigitalIrq );
	unsigned long func_code = pf_code & 0xFFFFFF;
	unsigned char dev_no;
	cell ret = -1;

	if( ( pf_code & pf_mask ) == pf_mask ) {	// GPIO(荞ݗL)̏ꍇ
		dev_no = pf_code & 0xF;
		if( dev_no < g_NumDigitalIn ) {
			func_code &= (~PinFuncDigitalIrq);		// @\R[h̐ݒ(荞ݖ)

			if( SetPinFuncCode( amx , pin_code , func_code ) == AMX_ERR_NONE ) {				// IOR[hPawnVM֏߂
				MAP_GPIOPinIntDisable( g_DI.Pins[ dev_no ].port , g_DI.Pins[ dev_no ].pin );	// GPIOPin荞݂𖳌ɂ
				ret = (pin_code << 24) | func_code ;											// IOR[h߂l֑
				g_GPIOPinIntRegistered[ dev_no ] = false;										// GPIOPin荞ݓo^tO
			}
		}
	}
	if( ret == (-1) ) amx_putstr("illegal I/O\n");

	return ret;
}

static cell AMX_NATIVE_CALL n_GPIOPortIntRegister(AMX *amx,const cell *params)
{
	unsigned char port_no = ( unsigned char )params[1];			// |[gԍ(1)̎擾
	unsigned long ulPort = g_ulGPIOBase[ port_no ];				// Y|[g̃x[XAhX擾
	int funcIdx;
	char *func_name;
	cell ret = -1;
	
	if( !g_GPIOPortIntRegistered[ port_no ] ) { 		// GPIO荞ݐݒ肪ĂȂꍇ
		amx_StrParam( amx , params[2] , func_name );	// Pawn̊荞݊֐(2)擾
		if( func_name != NULL) {
			if( amx_FindPublic( amx , func_name , &funcIdx) == AMX_ERR_NONE ) { // ֐ID擾
				g_GPIOIntHandlerId[ port_no ] = funcIdx;						// ֐IDۑ

				MAP_GPIOPinIntClear( ulPort , 0xFF );						// 荞݂NA
				MAP_IntDisable( g_ulGPIOInt[ port_no ] ); 					// GPIOPort荞݂~߂
				GPIOPortIntRegister( ulPort , GPIOIntHandler[port_no ] );	// GPIOPort荞݃nh̓o^
				g_GPIOPortIntRegistered[ port_no ] = true;					// GPIOPort荞ݓo^tO𗧂Ă
				ret = 1;
			}
			else {
				amx_putstr( func_name );
				amx_putstr(" not found\n");
				return amx_RaiseError( amx , AMX_ERR_NOTFOUND );	// sG[𔭐
			}
		}
	}
	return ret;
}

int amx_lm3s3748_GPIOInit(AMX *amx)
{
	unsigned char i;

	static AMX_NATIVE_INFO GPIO_Extension[] = 
	{
		{ "PinTypeGPIOOutput"   , n_PinTypeGPIOOutput },
		{ "PinTypeGPIOInput"    , n_PinTypeGPIOInput },
		{ "GPIOWrite"           , n_GPIOWrite },
		{ "GPIORead"            , n_GPIORead },
		{ "GPIOPortIntRegister" , n_GPIOPortIntRegister },
		{ "GPIOPinIntEnable"    , n_GPIOPinIntEnable } ,
		{ "GPIOPinIntDisable"   , n_GPIOPinIntDisable } ,
		{ NULL    , NULL }        // terminator
	};

	for( i=0 ; i<DIGITAL_IO_MAX; i++)	g_GPIOPinIntRegistered[i] = false;
	for( i=0 ; i<8; i++) g_GPIOPortIntRegistered[i] = false;

	return amx_Register(amx , GPIO_Extension , -1);
}

int amx_lm3s3748_GPIOCleanup(AMX *amx){
	unsigned char i;
	long status;

	g_NumDigitalOut = 0;
	g_NumDigitalIn  = 0;

	for( i=0 ; i<DIGITAL_IO_MAX; i++) {
		if( g_GPIOPinIntRegistered[i] ) {	// GPIOPin荞݂Lȏꍇ
			MAP_GPIOPinIntDisable( g_DI.Pins[ i ].port , g_DI.Pins[ i ].pin );	// GPIOPin荞݂𖳌ɂ
		}
		g_GPIOPinIntRegistered[ i ] = false;	// GPIOPin荞ݓo^tO
	}

	for( i=0 ; i<8; i++) {
		if( g_GPIOPortIntRegistered[i] )  {	// GPIOPort荞݂Lȏꍇ
			status = MAP_GPIOPinIntStatus( g_ulGPIOBase[i] , true );	// 荞ݏԂ擾
			MAP_GPIOPinIntClear( g_ulGPIOBase[i] , status );			// 荞݂NA
			MAP_IntDisable( g_ulGPIOInt[ i ] );							// GPIOPort荞݂~߂
		}
		g_GPIOPortIntRegistered[i] = false;	// GPIOPort荞ݓo^tO
	}
	return AMX_ERR_NONE;
}

extern AMX g_amx;	// PawnVMipawn_run.cŒ`j

// GPIOA荞݊֐
void PORTAIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTA_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTA_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTA ] ) {	// GPIOPortÅ荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTA << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTA ] );	// ΉID̊֐Ăяo(1)
	}
}

// GPIOB荞݊֐
void PORTBIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTB_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTB_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTB ] ) {	// GPIOPortB̊荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTB << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTB ] );	// ΉID̊֐Ăяo(1)
	}
}

// GPIOC荞݊֐
void PORTCIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTC_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTC_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTC ] ) {	// GPIOPortC̊荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTC << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTC ] );	// ΉID̊֐Ăяo(1)
	}
}

// GPIOD荞݊֐
void PORTDIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTD_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTD_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTD ] ) {	// GPIOPortD̊荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTD << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTD ] );	// ΉID̊֐Ăяo(1)
	}
}

// GPIOE荞݊֐
void PORTEIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTE_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTE_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTE ] ) {	// GPIOPortE̊荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTE << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTE ] );	// ΉID̊֐Ăяo(1)
	}
}

// GPIOF荞݊֐
void PORTFIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTF_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTF_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTF ] ) {	// GPIOPortF̊荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTF << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTF ] );	// ΉID̊֐Ăяo(1)
	}
}

// GPIOG荞݊֐
void PORTGIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTG_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTG_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTG ] ) {	// GPIOPortG̊荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTG << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTG ] );	// ΉID̊֐Ăяo(1)
	}
}

// GPIOH荞݊֐
void PORTHIntHandler(void)
{
	cell ret;

	long status = MAP_GPIOPinIntStatus( GPIO_PORTH_BASE , true );	// 荞ݏԂ擾
	MAP_GPIOPinIntClear( GPIO_PORTH_BASE , status );				// 荞݂NA

	if( g_GPIOPortIntRegistered[ PORTH ] ) {	// GPIOPortH̊荞݂o^Ăꍇ
		amx_Push(&g_amx , (PORTH << 8) | (status & 0xFF) );		// 1̃f[^PawnVMɓn
		amx_Exec(&g_amx , &ret, g_GPIOIntHandlerId[ PORTH ] );	// ΉID̊֐Ăяo(1)
	}
}

